// List.h
// Copyright (C) 2010 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _List_h_
#define _List_h_

#include <Basic/Assert.h>
#include <Basic/Debug.h>
#include <Basic/Memory.h>


template <typename T>
class ListItem
{
	public:

		T value;
		ListItem* const previous;
		ListItem* const next;


		ListItem() throw()
		:	previous((ListItem*)0),
			next((ListItem*)0)
		{
		}


		ListItem(const ListItem& list_item) throw()
		:	value(list_item.value),
			previous((ListItem*)0),
			next((ListItem*)0)
		{
		}


		ListItem(const T& value, ListItem* previous, ListItem* next) throw()
		:	value(value),
			previous(previous),
			next(next)
		{
		}


		ListItem& operator = (const ListItem& list_item) throw()
		{
			if (this != &list_item)
			{
				value = list_item.value;
			}
			return *this;
		}
};


template <typename T>
class List
{
	public:

		ListItem<T>* const first;
		ListItem<T>* const last;
		const int size;


		List() throw()
		:	first((ListItem<T>*)0),
			last((ListItem<T>*)0),
			size(0)
		{
		}


		List(const List& list) throw(Memory::Exception)
		:	first((ListItem<T>*)0),
			last((ListItem<T>*)0),
			size(0)
		{
			ListItem<T>* item = list.first;
			while (item)
			{
				AppendLast(item->value);
				item = item->next;
			}
		}


		~List() throw()
		{
			while (first)
			{
				ListItem<T>* next = first->next;
				delete first;
				*(ListItem<T>**)&first = next;
			}
		}


		List& operator = (const List& list) throw(Memory::Exception)
		{
			if (this != &list)
			{
				Clear();
				ListItem<T>* item = list.first;
				while (item)
				{
					AppendLast(item->value);
					item = item->next;
				}
			}
			return *this;
		}


		ListItem<T>* AppendFirst(ListItem<T>* new_item) throw(Memory::Exception)
		{
			Assert(new_item);

			*(ListItem<T>**)&new_item->next = first;
			if (first)
			{
				*(ListItem<T>**)&first->previous = new_item;
			}
			else
			{
				*(ListItem<T>**)&last = new_item;
			}
			*(ListItem<T>**)&first = new_item;
			++*(int*)&size;
			return first;
		}


		ListItem<T>* AppendLast(ListItem<T>* new_item) throw(Memory::Exception)
		{
			Assert(new_item);

			*(ListItem<T>**)&new_item->previous = last;
			if (last)
			{
				*(ListItem<T>**)&last->next = new_item;
			}
			else
			{
				*(ListItem<T>**)&first = new_item;
			}
			*(ListItem<T>**)&last = new_item;
			++*(int*)&size;
			return last;
		}


		ListItem<T>* AppendFirst(const T& value) throw(Memory::Exception)
		{
			ListItem<T>* new_item = new ListItem<T>(value, (ListItem<T>*)0, first);
			if (!new_item)
			{
				DebugPosition("Memory::Exception", (char*)0);
				throw Memory::exception;
			}
			if (first)
			{
				*(ListItem<T>**)&first->previous = new_item;
			}
			else
			{
				*(ListItem<T>**)&last = new_item;
			}
			*(ListItem<T>**)&first = new_item;
			++*(int*)&size;
			return first;
		}


		ListItem<T>* AppendLast(const T& value) throw(Memory::Exception)
		{
			ListItem<T>* new_item = new ListItem<T>(value, last, (ListItem<T>*)0);
			if (!new_item)
			{
				DebugPosition("Memory::Exception", (char*)0);		
				throw Memory::exception;
			}
			if (last)
			{
				*(ListItem<T>**)&last->next = new_item;
			}
			else
			{
				*(ListItem<T>**)&first = new_item;
			}
			*(ListItem<T>**)&last = new_item;
			++*(int*)&size;
			return last;
		}


		void Clear() throw()
		{
			while (first)
			{
				ListItem<T>* next = first->next;
				delete first;
				*(ListItem<T>**)&first = next;
			}
			*(ListItem<T>**)&last = (ListItem<T>*)0;
			*(int*)&size = 0;
		}


		void Delete(ListItem<T>* item) throw()
		{
			Assert(item);

			if (first == item)
			{
				*(ListItem<T>**)&first = item->next;
			}
			if (last == item)
			{
				*(ListItem<T>**)&last = item->previous;
			}
			if (item->previous)
			{
				*(ListItem<T>**)&item->previous->next = item->next;
			}
			if (item->next)
			{
				*(ListItem<T>**)&item->next->previous = item->previous;
			}
			delete item;
			--*(int*)&size;
		}


		ListItem<T>* Search(const T& value) throw()
		{
			register ListItem<T>* __restrict item = first;
			while (item)
			{
				if (item->value == value)
				{
					return item;
				}
				item = item->next;
			}
			return (ListItem<T>*)0;
		}
};

#endif
