// Set.h
// Copyright (C) 2010 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _Set_h_
#define _Set_h_

#include <Basic/Debug.h>
#include <Basic/Assert.h>
#include <Basic/Memory.h>


template <typename T>
class SetItem
{
	public:

		T value;
		SetItem* const previous;
		SetItem* const next;

		SetItem() throw()
		:	previous((SetItem*)0),
			next((SetItem*)0)
		{
		}

		SetItem(const SetItem& list_item) throw()
		:	value(list_item.value),
			previous((SetItem*)0),
			next((SetItem*)0)
		{
		}

		SetItem(const T& value, SetItem* previous, SetItem* next) throw()
		:	value(value),
			previous(previous),
			next(next)
		{
		}

		SetItem& operator = (const SetItem& list_item) throw()
		{
			if (this != &list_item)
			{
				value = list_item.value;
			}
			return *this;
		}
};


template <typename T>
class Set
{
	public:

		SetItem<T>* const first;
		SetItem<T>* const last;
		const int size;
	
		Set() throw()
		:	first((SetItem<T>*)0),
			last((SetItem<T>*)0),
			size(0)
		{
		}
	
		Set(const Set& list) throw(Memory::Exception)
		:	first((SetItem<T>*)0),
			last((SetItem<T>*)0),
			size(0)
		{
			SetItem<T>* item = list.first;
			while (item)
			{
				Append(item->value);
				item = item->next;
			}
		}
	
		~Set() throw()
		{
			while (first)
			{
				SetItem<T>* next = first->next;
				delete first;
				*(SetItem<T>**)&first = next;
			}
		}
	
		Set& operator = (const Set& list) throw(Memory::Exception)
		{
			if (this != &list)
			{
				Clear();
				SetItem<T>* item = list.first;
				while (item)
				{
					Append(item->value);
					item = item->next;
				}
			}
			return *this;
		}

		SetItem<T>* Append(const T& value) throw(Memory::Exception)
		{
			register SetItem<T>* search_item = first;
			while (search_item)
			{
				if (value == search_item->value)
				{
					return search_item;
				}
				if (value < search_item->value)
				{
					SetItem<T>* new_item = new SetItem<T>(value, search_item->previous, search_item);
					if (!new_item)
					{
						DebugPosition("Memory::Exception", (char*)0);		
						throw Memory::exception;
					}
					if (search_item->previous)
					{
						*(SetItem<T>**)&search_item->previous->next = new_item;
					}
					else
					{
						*(SetItem<T>**)&first = new_item;
					}
					*(SetItem<T>**)&search_item->previous = new_item;
					++*(int*)&size;
					return new_item;
				}
				search_item = search_item->next;
			}
			SetItem<T>* new_item = new SetItem<T>(value, last, (SetItem<T>*)0);
			if (!new_item)
			{
				DebugPosition("Memory::Exception", (char*)0);		
				throw Memory::exception;
			}
			if (last)
			{
				*(SetItem<T>**)&last->next = new_item;
			}
			else
			{
				*(SetItem<T>**)&first = new_item;
			}
			*(SetItem<T>**)&last = new_item;
			++*(int*)&size;
			return new_item;
		}

		void Clear() throw()
		{
			while (first)
			{
				SetItem<T>* next = first->next;
				delete first;
				*(SetItem<T>**)&first = next;
			}
			*(SetItem<T>**)&last = (SetItem<T>*)0;
			*(int*)&size = 0;
		}
	
		void Delete(SetItem<T>* item) throw()
		{
			Assert(item);

			if (first == item)
			{
				*(SetItem<T>**)&first = item->next;
			}
			if (last == item)
			{
				*(SetItem<T>**)&last = item->previous;
			}
			if (item->previous)
			{
				*(SetItem<T>**)&item->previous->next = item->next;
			}
			if (item->next)
			{
				*(SetItem<T>**)&item->next->previous = item->previous;
			}
			delete item;
			--*(int*)&size;
		}

		void Delete(const T& value) throw()
		{
			SetItem<T>* item = Search(value);
			Assert(item);
			Delete(item);
		}

		SetItem<T>* Search(const T& value) throw()
		{
			register SetItem<T>* __restrict item = first;
			while (item)
			{
				if (value == item->value)
				{
					return item;
				}
				if (value < item->value)
				{
					break;
				}
				item = item->next;
			}
			return (SetItem<T>*)0;
		}
};

#endif
