// PowerIteration.h
// Copyright (C) 2011 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _PowerIteration_h_
#define _PowerIteration_h_

#include <Basic/Log.h>
#include <Basic/Memory.h>
#include <Container/CSRMatrix.h>
#include <Container/Vector.h>
#include <math.h>


template <typename T>
void PowerIteration(const CSRMatrix<T>& A, Vector<T>& v, T& lambda, T tolerance, int max_steps) throw(Memory::Exception)
{
	Vector<T> vt(v.size);
	int N = A.rows;
	for (int s = 1; s <= max_steps; ++s)
	{
		#pragma omp parallel for default(shared) schedule(guided)
		for (int i = 1; i <= N; ++i)
		{
			T sum = 0;
			int k_max = A.count[i];
			for (int k = 1; k <= k_max; ++k)
			{
				sum += A.entry[i][k]*v.entry[A.index[i][k]];
			}
			vt.entry[i] = sum;
		}
		T vt_max = 0;
		for (int i = 1; i <= N; ++i)
		{
			T vt = fabs(vt.entry[i]);
			if (vt_max < vt)
			{
				vt_max = vt;
			}
		}
		#pragma omp parallel for default(shared)
		for (int i = 1; i <= N; ++i)
		{
			v.entry[i] = vt.entry[i]/vt_max;
		}
	}
}


template <typename T>
void PowerMethod(const CSRMatrix<T>& A, Vector<T>& v, T& lambda, T tolerance, int max_steps) throw(Memory::Exception)
{
	v.Fill(1);
}


void InversePowerIteration(const CSRMatrix<T>& L, const CSRMatrix<T>& U, Vector<T>& v, T& lambda, T tolerance, int max_steps) throw(Memory::Exception)
{
	Vector<T> wt(v.size);
	Vector<T> vt(v.size);
	int N = A.rows;
	for (int s = 1; s <= max_steps; ++s)
	{
		LowerTriangularSystem(L, w, v);
		UpperTriangularSystem(U, vt, w);

		T vt_max = 0;
		for (int i = 1; i <= N; ++i)
		{
			T vt = fabs(vt.entry[i]);
			if (vt_max < vt)
			{
				vt_max = vt;
			}
		}
		#pragma omp parallel for default(shared)
		for (int i = 1; i <= N; ++i)
		{
			v.entry[i] = vt.entry[i]/vt_max;
		}
	}
}


void InversePowerMethod(const CSRMatrix<T>& A, Vector<T>& v, T& lambda, T tolerance, int max_steps) throw(Memory::Exception)
{
	v.Fill(1);
}

#endif
