// main.mpi.cpp
// Copyright (C) 2011 Miguel Vargas (miguel.vargas@gmail.com)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <Basic/Log.h>
#include <Basic/Macros.h>
#include <File/File.h>
#include <FiniteElement/Assembler.h>
#include <FiniteElement/Mesh.h>
#include <Substructuring/SchurComplement.h>

#ifdef WIN32
	#define PIPE_PREFIX "\\\\.\\pipe\\"
#else
	#define PIPE_PREFIX "/tmp/"
#endif

#define VERSION "1.0.0"

// Default environment variables values
#define LOG_LEVEL                2
#define SOLVER_THREADS           1
#define SUBSTRUCTURING_THREADS   1
#define SUBSTRUCTURING_TOLERANCE 1e-5


enum Command
{
	command_end              = 0,
	command_set_connectivity = 1,
	command_fill_K           = 2,
	command_set_Ke           = 3,
	command_set_all_Ke       = 4,
	command_set_u            = 5,
	command_set_f            = 6,
	command_set_fixed        = 7,
	command_solver_init      = 8,
	command_solver_run       = 9
};


int main(int argc, char** argv)
{
	if ((argc != 1) && (argc != 3))
	{
		fprintf(stderr, "Invalid number of arguments. Use:\n  %s [<data_pipe_name> <results_pipe_name>]\n", argv[0]);
		fputs("In Windows systems, pipe names must have the form: \\\\.\\pipe\\[pipe_name]\n", stderr);
		return 1;
	}

	try
	{
		int solver_threads = GetEnvInteger(SOLVER_THREADS);
		int substructuring_threads = GetEnvInteger(SUBSTRUCTURING_THREADS);
		double substructuring_tolerance = GetEnvFloat(SUBSTRUCTURING_TOLERANCE);

		// MPI object
		MPI mpi(argc, argv);
		if (mpi.size < 3)
		{
			if (mpi.rank == 0)
			{
				fputs("[Error] Invalid number processes to start, minimum 3 processes should be started.\n", stderr);
			}
			return 1;
		}

		if (mpi.rank == 0) // Master
		{
			// Set log level from environment
			log_level = GetEnvInteger(LOG_LEVEL);

			int partitions_count = mpi.size - 1;

			Log(1, "FEMSolver.MPI --------------------------------------------------------");
			Log(1, "-Version:    %s", VERSION);
			Log(1, "-Partitions: %i", partitions_count);

			// Create and open pipes
			Log(1, "Create pipes ---------------------------------------------------------");
			File data_pipe;
			const char* data_path = (argc == 3) ? argv[1] : PIPE_PREFIX "FEMData";
			Log(1, "Data pipe: %s", data_path);
			File result_pipe;
			const char* result_path = (argc == 3) ? argv[2] : PIPE_PREFIX "FEMResult";
			Log(1, "Result pipe: %s", result_path);
			#pragma omp parallel sections num_threads(2)
			{
				#pragma omp section
				{
					data_pipe.NamedPipe(data_path, named_pipe_read_only);
				}
				#pragma omp section
				{
					result_pipe.NamedPipe(result_path, named_pipe_write_only);
				}
			}

			// Geometry data
			Mesh mesh;
			int degrees_of_freedom = 0;

			// Elemental matrices data
			int N = 0;
			Vector<Matrix<double> > Ke;

			// Problem data
			int M = 0;
			Vector<double> global_u;
			Vector<double> global_f;
			Vector<bool> global_fixed;

			for (bool go = true; go; )
			{
				char command;
				data_pipe.Read(command);
				switch ((Command)command)
				{
					case command_end:
					{
						result_pipe.Close();
						data_pipe.Close();
						go = false;
						continue;
					}
					case command_set_connectivity:
					{
						int nodes_count;
						int elements_count;
						int element_type;
						int nodes_per_element;

						data_pipe.Read(&nodes_count, 1);
						data_pipe.Read(&elements_count, 1);
						data_pipe.Read(&element_type, 1);
						data_pipe.Read(&nodes_per_element, 1);
						data_pipe.Read(&degrees_of_freedom, 1);

						mesh.Resize((ShapeType)element_type, nodes_per_element, elements_count, nodes_count);
						data_pipe.Read(mesh.connectivity.data, elements_count*nodes_per_element);
						mesh.PrintInfo();

						M = nodes_count*degrees_of_freedom;
						global_u.Resize(M);
						global_f.Resize(M);
						global_fixed.Resize(M);

						N = nodes_per_element*degrees_of_freedom;
						Ke.Resize(elements_count);
						for (int e = 1; e <= elements_count; ++e)
						{
							Ke.entry[e].Resize(N, N);
						}
						continue;
					}
					case command_fill_K:
					{
						continue;
					}
					case command_set_Ke:
					{
						int element_id;
						data_pipe.Read(&element_id, 1);
						data_pipe.Read(Ke.entry[element_id].data, N*N);
						continue;
					}
					case command_set_all_Ke:
					{
						for (int element_id = 1; element_id <= mesh.elements_count; ++element_id)
						{
							data_pipe.Read(Ke.entry[element_id].data, N*N);
						}
						continue;
					}
					case command_set_u:
					{
						data_pipe.Read(global_u.data, global_u.size);
						continue;
					}
					case command_set_f:
					{
						data_pipe.Read(global_f.data, global_f.size);
						continue;
					}
					case command_set_fixed:
					{
						data_pipe.Read(global_fixed.data, global_fixed.size);
						continue;
					}
					case command_solver_init:
					{
						continue;
					}
					case command_solver_run:
					{
						continue;
					}
					default:
					{
						Log(1, "Invalid command %i", command);
						result_pipe.Close();
						data_pipe.Close();
						go = false;
						continue;
					}
				}
			}
		}
		else // rank != 0
		{
			log_level = 0; // Prevents slaves logs to appear

			SchurComplementSlave<double> schur_complement_slave(mpi, solver_threads);
			schur_complement_slave.ReceiveSystemOfEquations();
			schur_complement_slave.Solver();
			schur_complement_slave.CompleteSolution();
			schur_complement_slave.ReportMemoryPeakUsage();
		}
	}
	catch (Exception&)
	{
		DebugPosition("Catch", (char*)0);
		fputs("[Error] Fatal exception\n", stderr);
	}

	if (Memory::current_usage != 0)
	{
		fprintf(stderr, "[Error] Memory leak: %lu bytes\n", Memory::current_usage);
	}

	return 0;
}
